/*
 * 68k.c: 68000 CPU emulation abstraction interface
 */

/*
 * For Starscream debugger, #define STARSCREAM_DEBUG, also make sure to
 * modify the Makefile to create star/cpudebug.o
 */

/* #define STARSCREAM_DEBUG */

#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include "68k.h"
#include "gen68k.h"
#include "genital.h"
#include "input.h"
#ifdef STARSCREAM_DEBUG
        #include "star/cpudebug.h"
#endif
#include "star/starcpu.h"
#include "vdp.h"
#include "vga.h"


/* CPU functions */
int             (*CPURun)(int);
int             (*CPUInterrupt)(int);

/* internal handler prototypes */
unsigned        Gen68KReadMemByte(unsigned);
unsigned        Gen68KReadMemWord(unsigned);
void            Gen68KWriteMemByte(unsigned, unsigned);
void            Gen68KWriteMemWord(unsigned, unsigned);
unsigned        Gen68KFakeRead(unsigned);
void            Gen68KFakeWrite(unsigned, unsigned);
unsigned        StarscreamReadRAMByte(unsigned);
unsigned        StarscreamReadRAMWord(unsigned);
void            StarscreamWriteRAMByte(unsigned, unsigned);
void            StarscreamWriteRAMWord(unsigned, unsigned);
unsigned        StarscreamFakeRead(unsigned);
void            StarscreamFakeWrite(unsigned, unsigned);

int             cpu_type = DEFAULT_CPU;         /* current CPU emulator */


/*
 * Genital68K memory map
 */

struct GENITAL68K_INSTRUCTIONREGION     gen68k_fetch_instructions[] =
{
        { 0x000000, 0x3fffff, NULL },
        { -1, -1, NULL }
};
struct GENITAL68K_DATAREGIONREAD        gen68k_read_data_byte[] =
{
        { 0x000000, 0x3fffff, Gen68KReadMemByte },
        { 0x400000, 0x9fffff, Gen68KFakeRead },
        { 0xa00000, 0xbfffff, IOReadByte },
        { 0xc00000, 0xdfffff, VDPReadByte },
        { 0xe00000, 0xffffff, Gen68KReadMemByte },
        { -1, -1, NULL }
};
struct GENITAL68K_DATAREGIONREAD        gen68k_read_data_word[] =
{
        { 0x000000, 0x3fffff, Gen68KReadMemWord },
        { 0x400000, 0x9fffff, Gen68KFakeRead },
        { 0xa00000, 0xbfffff, IOReadWord },
        { 0xc00000, 0xdfffff, VDPReadWord },
        { 0xe00000, 0xffffff, Gen68KReadMemWord },
        { -1, -1, NULL }
};
struct GENITAL68K_DATAREGIONWRITE       gen68k_write_data_byte[] =
{
        { 0x000000, 0x9fffff, Gen68KFakeWrite },
        { 0xa00000, 0xbfffff, IOWriteByte },
        { 0xc00000, 0xdfffff, VDPWriteByte },
        { 0xe00000, 0xffffff, Gen68KWriteMemByte },
        { -1, -1, NULL }
};
struct GENITAL68K_DATAREGIONWRITE       gen68k_write_data_word[] =
{
        { 0x000000, 0x9fffff, Gen68KFakeWrite },
        { 0xa00000, 0xbfffff, IOWriteWord },
        { 0xc00000, 0xdfffff, VDPWriteWord }, 
        { 0xe00000, 0xffffff, Gen68KWriteMemWord },
        { -1, -1, NULL }
};


/*
 * Starscream memory map (all NULL items initialized by CPUInit())
 */

struct STARSCREAM_PROGRAMREGION star_fetch_instructions[] =
{
        { 0x000000, 0x3fffff, (unsigned) NULL },
        { 0xff0000, 0xffffff, (unsigned) NULL },
        { 0xff000000, 0xff3fffff, (unsigned) NULL },
        { 0xfff00000, 0xffffffff, (unsigned) NULL },
        { -1, -1, NULL }
};
struct STARSCREAM_DATAREGION    star_read_data_byte[] =
{
        { 0x000000, 0x3fffff, NULL, (void *) NULL },
        { 0x400000, 0x9fffff, StarscreamFakeRead, NULL },
        { 0xa00000, 0xbfffff, IOReadByte, NULL },
        { 0xc00000, 0xdfffff, VDPReadByte, NULL },
        { 0xe00000, 0xffffff, StarscreamReadRAMByte, NULL },
        { -1, -1, NULL, NULL }
};
struct STARSCREAM_DATAREGION    star_read_data_word[] =
{
        { 0x000000, 0x3fffff, NULL, (void *) NULL },
        { 0x400000, 0x9fffff, StarscreamFakeRead, NULL },
        { 0xa00000, 0xbfffff, IOReadWord, NULL },
        { 0xc00000, 0xdfffff, VDPReadWord, NULL },
        { 0xe00000, 0xffffff, StarscreamReadRAMWord, NULL },
        { -1, -1, NULL, NULL }
};
struct STARSCREAM_DATAREGION    star_write_data_byte[] =
{
        { 0x000000, 0x9fffff, StarscreamFakeWrite, NULL },
        { 0xa00000, 0xbfffff, IOWriteByte, NULL },
        { 0xc00000, 0xdfffff, VDPWriteByte, NULL },
        { 0xe00000, 0xffffff, StarscreamWriteRAMByte, NULL },
        { -1, -1, NULL, NULL }
};
struct STARSCREAM_DATAREGION    star_write_data_word[] =
{
        { 0x000000, 0x9fffff, StarscreamFakeWrite, NULL },
        { 0xa00000, 0xbfffff, IOWriteWord, NULL },
        { 0xc00000, 0xdfffff, VDPWriteWord, NULL }, 
        { 0xe00000, 0xffffff, StarscreamWriteRAMWord, NULL },
        { -1, -1, NULL, NULL }
};


/*
 * Genital68K specific function handlers
 */

unsigned Gen68KReadMemByte(unsigned addr)
{
        if (addr < 0x400000)
                return genesis.rom[addr ^ 1];
        else if (addr >= 0xe00000)
                return genesis.ram[(addr ^ 1) & 0xffff];
        return 0;        
}
        
unsigned Gen68KReadMemWord(unsigned addr)
{
        if (addr < 0x400000)
                return *((unsigned short *) (genesis.rom + (addr & 0x3ffffe)));
        else if (addr >= 0xe00000)
                return *((unsigned short *) (genesis.ram + (addr & 0xfffe)));
        return 0;
}

void Gen68KWriteMemByte(unsigned addr, unsigned data)
{
        if (addr >= 0xe00000)
                genesis.ram[(addr ^ 1) & 0xffff] = data;                
}

void Gen68KWriteMemWord(unsigned addr, unsigned data)
{
        if (addr >= 0xe00000)
                *((unsigned short *) (genesis.ram + (addr & 0xfffe))) = data;
}

unsigned Gen68KFakeRead(unsigned addr)
{
        return 0;
}

void Gen68KFakeWrite(unsigned addr, unsigned data)
{       
}

int Gen68KRun(int cycles)
{
        Genital68KRun(cycles);
        return 0;
}

int Gen68KInterrupt(int int_type)
{
        return 0;
}


/*
 * Starscream specific function handlers
 */

unsigned StarscreamReadRAMByte(unsigned address)
{
        return genesis.ram[(address ^ 1) & 0xffff];
}
        
unsigned StarscreamReadRAMWord(unsigned address)
{
        return *((unsigned short *) (genesis.ram + (address & 0xfffe)));
}

void StarscreamWriteRAMByte(unsigned address, unsigned data)
{
        genesis.ram[(address ^ 1) & 0xffff] = data;
}

void StarscreamWriteRAMWord(unsigned address, unsigned data)
{
        *((unsigned short *) (genesis.ram + (address & 0xfffe))) = data;
}

unsigned StarscreamFakeRead(unsigned address)
{
        return 0;
}

void StarscreamFakeWrite(unsigned address, unsigned data)
{       
}

int StarscreamRun(int cycles)
{
        s68000tripOdometer();           /* reset odometer */
        s68000exec(cycles);
        return 0;
}

int StarscreamInterrupt(int int_type)
{
        if (int_type == HINTERRUPT)
                s68000interrupt(4, -1);
        else if (int_type == VINTERRUPT)
                s68000interrupt(6, -1);
        else if (int_type == EXTINTERRUPT)
                s68000interrupt(2, -1);
        return 0;
}
        
void StarscreamResetHandler()
{
}


/*
 *  General 68000 functions
 */

void CPUDebug()
{
        int     video_mode;

#ifdef STARSCREAM_DEBUG
        if (cpu_type == STARSCREAM_CPU)
        {
                video_mode = GraphicsGetMode();
                GraphicsSetMode(TEXTMODE);
                KeyboardDeinit();
                cpudebug_interactive(1, NULL, NULL, NULL, NULL);
                KeyboardInit();
                GraphicsSetMode(video_mode);
        }
#endif
}
                
unsigned CPUGetPC()
{
        if (cpu_type == STARSCREAM_CPU)
                return s68000readPC();
        return 0;
}

void CPUSetPC(unsigned data)
{
        if (cpu_type == STARSCREAM_CPU)
                s68000context.pc = data;
}

unsigned CPUGetSR()
{
        if (cpu_type == STARSCREAM_CPU)
                return s68000context.sr;
        return 0;
}

void CPUSetSR(unsigned data)
{
        if (cpu_type == STARSCREAM_CPU)
                s68000context.sr = data;
}

unsigned CPUGetA(int i)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                return s68000context.areg[i];
        return 0;
}

void CPUSetA(int i, unsigned data)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                s68000context.areg[i] = data;
}        

unsigned CPUGetD(int i)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                return s68000context.dreg[i];
        return 0;
}

void CPUSetD(int i, unsigned data)
{
        i &= 7;
        if (cpu_type == STARSCREAM_CPU)
                s68000context.dreg[i] = data;
}        


void CPUReset()
{
        if (cpu_type == STARSCREAM_CPU)
                s68000reset();
}

int CPUGetType()
{
        return cpu_type;
}

void CPUInit(int type)
{
        if (type == GENITAL68K_CPU)             /* Genital68K core */
        {
                cpu_type = GENITAL68K_CPU;

                gen68k_fetch_instructions[0].offset = genesis.rom;

                /* initialize the default context provided by Genital68K */
                genital68kcontext.fetch = gen68k_fetch_instructions;
                genital68kcontext.read_byte = gen68k_read_data_byte;
                genital68kcontext.read_word = gen68k_read_data_word;
                genital68kcontext.write_byte = gen68k_write_data_byte;
                genital68kcontext.write_word = gen68k_write_data_word;

                /* set up handlers */
                CPURun = Gen68KRun;
                CPUInterrupt = Gen68KInterrupt;

                /* Genital68K requires bytes to be swapped */
                SwapBuffer(genesis.rom, 4194304);

                /* initialize Genital68K */
                if (Genital68KInit(&genital68kcontext))
                {
                        printf("Genital: Failed to initialize 68K\n");
                        exit(1);
                }
                Genital68KReset();
        }
        else if (type == STARSCREAM_CPU)        /* Starscream core */
        {
                cpu_type = STARSCREAM_CPU;

                /* set some context pointers */
                star_fetch_instructions[0].offset
                                = (unsigned) genesis.rom - 0x000000;
                star_fetch_instructions[1].offset
                                = (unsigned) genesis.ram - 0xff0000;
                star_fetch_instructions[2].offset
                                = (unsigned) genesis.rom - 0xff000000; //c
                star_fetch_instructions[3].offset
                                = (unsigned) genesis.ram - 0xfff00000;
                star_read_data_byte[0].userdata = (void *) genesis.rom;
                star_read_data_word[0].userdata = (void *) genesis.rom;

                /* initialize the default context provided by Starscream */
                s68000context.fetch = star_fetch_instructions;
                s68000context.readbyte = star_read_data_byte;
                s68000context.readword = star_read_data_word;
                s68000context.writebyte = star_write_data_byte;
                s68000context.writeword = star_write_data_word;
                s68000context.s_fetch = star_fetch_instructions;
                s68000context.s_readbyte = star_read_data_byte;
                s68000context.s_readword = star_read_data_word;
                s68000context.s_writebyte = star_write_data_byte;
                s68000context.s_writeword = star_write_data_word;
                s68000context.u_fetch = star_fetch_instructions;
                s68000context.u_readbyte = star_read_data_byte;
                s68000context.u_readword = star_read_data_word;
                s68000context.u_writebyte = star_write_data_byte;
                s68000context.u_writeword = star_write_data_word;
                s68000context.resethandler = StarscreamResetHandler;

                /* set up handlers */
                CPURun = StarscreamRun;
                CPUInterrupt = StarscreamInterrupt;

                /* Starscream requires bytes to be swapped */
                SwapBuffer(genesis.rom, 4194304);

                /* initialize Starscream */
                s68000init();
                if (s68000reset())
                {
                        printf("Genital: Starscream failed to initialize\n");
                        exit(1);
                }
        }
        else
        {
                printf("Genital: Unknown 68000 emulator type: %d\n", type);
                exit(1);
        }
}

